"""
$RCSfile: ExternalFile.py,v $

Implements ExternalFile class as a Zope product
ExternalFile encapsulates a file in the filesystem for
use as a Zope object.

Author: <a href="mailto:cstrong@arielpartners.com">Craeg Strong</a>
Release: 1.0
"""

__cvstag__  = '$Name:  $'[6:-2]
__date__    = '$Date: 2002/06/09 16:05:49 $'[6:-2]
__version__ = '$Revision: 1.71 $'[10:-2]

#
# Local modules
#
from FileUtils import copy_file
from ProductProperties import getMimeTypeList, getMimeTypeFromFile, getDisplayTypeFromMimeType

#
# Add our base classes.  Derived classes will get this stuff automatically
#
from OFS.SimpleItem import SimpleItem  # The SimpleItem module is a Zope builtin that provides
                                       # cut/paste, management views, webdav/ftp, undo, ownership,
                                       # traversal, acquisition, and persistence
from Products.ZCatalog.CatalogAwareness import CatalogAware # Reqd for ZCatalog support
from OFS.PropertyManager import PropertyManager  # Reqd for proper property mgmt
from IExternalFile import IExternalFile

# Zope builtins
from Globals import DTMLFile, MessageDialog
import Globals               # InitializeClass (security stuff)
import OFS                   # we grab his _property settings
from DateTime.DateTime import DateTime
from webdav.common import rfc1123_date
from AccessControl import ClassSecurityInfo

# Python builtins
from cStringIO import StringIO
import mimetypes, os, os.path, string

################################################################
# Container Methods
################################################################

def manage_add(self,
               id, title, description,
               target_filepath, basedir='', REQUEST=None):
    """

    Factory method to actually create an instance of ExternalFile.
    This method will be overridden in classes that inherit from
    ExternalFile.  This method assumes all parameters are correct (it
    does no error checking).  It is called from CreationDialog.py once
    all of the confirmation and error checking steps have been taken.
    
    You should call this method directly if you are creating an
    instance of ExternalFile programmatically and have 'vetted' all of
    your parameters for correctness.

    """
    self._setObject(id, ExternalFile(id, title, description,
                                     os.path.join(basedir,target_filepath)))
    self._getOb(id).reindex_object()

def manage_add_with_upload(self,
                           id, title, description,
                           target_filepath, upload_file, basedir='',
                           REQUEST=None):
    """

    Factory method to actually create an instance of ExternalFile.
    This method will be overridden in classes that inherit from
    ExternalFile.  This method includes an upload_file parameter.
    This method assumes all parameters are correct (it does no error
    checking).

    You should call this method directly if you are creating an
    instance of ExternalFile programmatically via an uploaded file and
    have 'vetted' all of your parameters for correctness.

    """
    fully_resolved_target_filepath = os.path.join(basedir,target_filepath)    

    copy_file(upload_file, fully_resolved_target_filepath)
    
    self._setObject(id, ExternalFile(id, title, description,
                                     fully_resolved_target_filepath))
    self._getOb(id).reindex_object()

################################################################
# ExternalFile class
################################################################

class ExternalFile(CatalogAware,    # ZCatalog support
                   PropertyManager, # Property support
                   SimpleItem):

    """
    ExternalFile encapsulates a file in the filesystem as a Zope object.
    """

    edit_externalfile_formpart    = DTMLFile('dtml/edit_externalfile_formpart', globals())
    
    meta_type = 'External File' # This is the name Zope will use for
                                # the Product in the "addProduct" list

    __implements__ = IExternalFile

    behave_like_list    = [ 'DTMLDocument', 'DTMLMethod', 'File', 'Image'  ]

    # This tuple defines our properties
    _properties = (
        {'id':'title',          'type':'string',    'mode': 'w'},
        {'id':'description',    'type':'string',    'mode': 'w'},
        {'id':'filepath',       'type':'string',    'mode': '' },
        {'id':'content_type',   'type':'selection', 'mode': 'w', 'select_variable':'getMimeTypes', },
        {'id':'behave_like',    'type':'selection', 'mode': 'w', 'select_variable':'behave_like_list', },        
        )
    
    # This tuple defines a dictionary for each tab in the management interface
    # label = label of tab, action = url it links to
    manage_options = (
        {'label':'Edit', 'action': 'manage_editForm', 'help': ('ExternalFile', 'edit.stx') },
        {'label':'View', 'action': 'manage_view',     'help': ('ExternalFile', 'view.stx') },
        ) + OFS.PropertyManager.PropertyManager.manage_options + OFS.SimpleItem.SimpleItem.manage_options

    _security = ClassSecurityInfo()

    # set security for the object itself, e.g. if it is accessed in DTML code
    # This line is REQUIRED to allow access to External Files by the public (ariel DTDs, etc.)
    _security.declareObjectPublic()

    def __init__(self, id, title, description, filepath):
        """
        Initialize a new instance of ExternalFile
        """
        self.id          = id       # This is the OID translated by Zope into a URL
        self.title       = title
        self.description = description
        self.filepath    = filepath # This is the full pathname of the file
        self.__version__ = __version__

        self.content_type = getMimeTypeFromFile(filepath)

        displaytype = getDisplayTypeFromMimeType(self.content_type)
        
        #
        # By default, behave like a DTMLDocument, unless this is an
        # image, in which case, we need to behave like an Image, or
        # binary, in which case we need to behave like a File.
        #
        if displaytype == 'image':
            self.behave_like = 'Image'
        elif displaytype == 'binary':
            self.behave_like = 'File'
        else:
            self.behave_like = 'DTMLDocument'

    # innocuous methods should be declared public
    _security.declarePublic('getMimeTypes')
    def getMimeTypes(self):
        "return list of available mime types."
        return getMimeTypeList()

    # innocuous methods should be declared public
    _security.declarePublic('getSelf')
    def getSelf(self):
        "Return this object. For use in DTML scripts"
        return self.aq_chain[0]
    
    ################################################################
    # Methods implementing the IExternalFile interface below
    ################################################################
    
    # protect these b/c a malicious hacker could get info about the filesystem
    _security.declareProtected('View',  'getContentType')
    def getContentType(self):
        """

        Returns the mimetype of the file.

        """
        return self.content_type

    # protect these b/c a malicious hacker could get info about the filesystem
    _security.declareProtected('View',  'getDisplayType')
    def getDisplayType(self):
        """

        Returns 'ascii', 'image', or 'binary' according to the
        mime_type_map defaults to 'binary' if no match is found.

        """
        return getDisplayTypeFromMimeType(self.content_type)

    # protect these b/c a malicious hacker could get info about the filesystem
    _security.declareProtected('View',  'getFilepath')
    def getFilepath(self, REQUEST):
        """

        Return the full pathname of the external file to which we are
        referring.  The REQUEST parameter encapsulates information
        about the environment which can be used by subclasses
        overriding this method.

        """
        return self.filepath


    _security.declareProtected('View', 'getFileModTime')
    def getFileModTime(self, REQUEST):
        """
        
        Return the last modification date of the external file to
        which we are referring.  The value is returned as number of
        seconds since the epoch in UTC (see python time module).  The
        REQUEST parameter encapsulates information about the
        environment which can be used by subclasses overriding this
        method.
        
        """
        path = self.getFilepath(REQUEST)
        return os.path.getmtime(path)

    _security.declareProtected('View', 'getFileModTime')
    def getModTime(self, REQUEST):
        """
        
        Return the last modification date of the external file to
        which we are referring, or of our metadata which is stored
        internally in the ZODB, whichever is more recent (greater).
        The value is returned as number of seconds since the epoch in
        UTC (see python time module).  The REQUEST parameter
        encapsulates information about the environment which can be
        used by subclasses overriding this method.
        
        """
        tempTime = self.getFileModTime(REQUEST)
        if self._p_mtime > tempTime:
            return self._p_mtime
        else:
            return tempTime

    _security.declareProtected('View', 'getContents')
    def getContents(self, REQUEST, RESPONSE=None):
        """

        Return the contents of this external file as a string.  The
        REQUEST parameter encapsulates information about the
        environment which can be used by subclasses overriding this
        method.
        
        """
        #
        # if file does not exist, should we throw an exception?
        # right now, we return the error message as if it
        # were the contents of the file.
        #
        # We need to do something better! cks 10/31/2001
        #
        #print "getContents of:",self.getFilepath(REQUEST)
        try:
            data = StringIO()
            path = self.getFilepath(REQUEST)
            copy_file(path, data)
        except IOError, value:
            return value
        else:
            return data.getvalue()

    _security.declareProtected('Change External Files', 'setContents')
    def setContents(self, newcontents, REQUEST):
        """
        
        Overwrite the previous contents of this file with the
        newcontents string.  The REQUEST parameter encapsulates
        information about the environment which can be used by
        subclasses overriding this method.

        """
        data = StringIO(newcontents)
        copy_file(data, self.getFilepath(REQUEST))
        # update ZCatalog
        self.reindex_object()
        # No need to set dirty flag, Zope object itself has not actually changed
        # self._p_changed = 1  # dirty flag, triggers persistence to update the object

    ################################################################
    # Standard Zope stuff
    ################################################################

    # call __call__() method on DTMLFile for our edit form
    # globals() is standard python for passing the current namespace
    _security.declareProtected('Change External Files', 'manage_editForm')
    manage_editForm = DTMLFile('dtml/edit',globals())

    # call __call__() method on DTMLFile for our view form
    # globals() is standard python for passing the current namespace
    _security.declareProtected('View', 'manage_view')
    manage_view           = DTMLFile('dtml/view',globals())

    _security.declarePublic('__len__')
    def __len__(self):
        "...unsure if this is required.  See get_size()"
        if os.path.exists(self.filepath):
            return os.stat(self.filepath)[6]
        else:
            return 0

    _security.declarePublic('get_size')
    def get_size(self):
        """

        Return the length of the underlying file.  This method is
        REQUIRED or FTP won't work.

        """
        if os.path.exists(self.filepath):
            return os.stat(self.filepath)[6]
        else:
            return 0 

    _security.declareProtected('Access contents information', 'index_html')
    # next line is not strictly necessary, Access contents info is Anonymous by default
    # as opposed to all other protections that get Manager Role by default
    _security.setPermissionDefault('Access contents information', ('Anonymous')) 
    def index_html(self, REQUEST):
        """

        Render the document/method/Image/File/etc appropriately.  Note
        that this method is called for ExternalFile when it is
        emulating a DTMLMethod, even though a DTMLMethod itself never
        calls index_html.

        """
        return self.__call__(None, REQUEST)

    # This is used for verification messages like "File Changed"
    # when you hit the button in the edit form.
    manage_header_message = DTMLFile('dtml/manage_header_message', globals())

    _security.declareProtected('Access contents information', '__str__')
    def __str__(self, REQUEST=None):
        """get contents as a string"""

        return self.__call__(None, REQUEST)
        
    _security.declareProtected('Access contents information', '__call__')
    def __call__(self, client=None, REQUEST=None):
        """

        Render the document given a client object, REQUEST mapping,
        Response, and key word arguments.

        """

        if self.behave_like == 'File':
            resultFile = OFS.Image.File(self.id, self.title,
                                        self.getContents(REQUEST),
                                        self.content_type)
            resultFile._p_mtime     = self.getModTime(REQUEST)
            resultFile.absolute_url = self.absolute_url
            resultFile.content_type = self.content_type
            if REQUEST:
                resp = REQUEST['RESPONSE'] 
                return resultFile.index_html(REQUEST, resp)
            else:
                return resultFile.__str__()

        if self.behave_like == 'Image':
            resultImage = OFS.Image.Image(self.id, self.title,
                                          self.getContents(REQUEST),
                                          self.content_type)
            resultImage._p_mtime     = self.getModTime(REQUEST)
            resultImage.absolute_url = self.absolute_url
            resultImage.content_type = self.content_type
            if REQUEST:
                resp = REQUEST['RESPONSE'] 
                return resultImage.index_html(REQUEST, resp)
            else:
                return resultImage.__str__()
            
        elif self.behave_like == 'DTMLDocument':
            me = self.getSelf()
            resultDocument = OFS.DTMLDocument.DTMLDocument(
                self.getContents(REQUEST))

            resultDocument.id           = self.id
            resultDocument.title        = self.title
            resultDocument.content_type = self.content_type            
            return resultDocument(me, REQUEST)
        
        else: # DTMLMethod
            myContainer = self.aq_chain[1]
            resultDocument = OFS.DTMLMethod.DTMLMethod(
                self.getContents(REQUEST))
            
            resultDocument.id           = self.id
            resultDocument.title        = self.title
            resultDocument.content_type = self.content_type            
            return resultDocument.__of__(myContainer)(myContainer, REQUEST)

    _size_changes = {
        # directive: (colDelta, rowDelta)
        'Narrower': (-5,  0),
        'Wider':    (5,   0),
        'Taller':   (0,   5),
        'Shorter':  (0,  -5),
        }

    _minCols = 40
    _minRows = 5

    _security.declareProtected('Change External Files', 'manage_editAction')
    def manage_editAction(self, externalfile_content = None,
                          SUBMIT = 'Cancel Changes',
                          edit_textarea_cols = '80',
                          edit_textarea_rows = '40',
                          upload_file = None,
                          REQUEST=None):
        """
        
        Called in response to edit.dtml form submit, to save changes,
        cancel changes, replace the contents of the file with uploaded
        contents, or change the size of the edit textarea form
        control.
        
        """
        if SUBMIT == 'Save Changes':
            self.setContents(externalfile_content, REQUEST)
            message = 'Content changed. (%s)' % DateTime().ISO()
            return self.manage_editForm(header_message=message)

        elif SUBMIT == 'Cancel Changes':
            message = 'Content changes cancelled.'
            return self.manage_editForm(externalfile_content,
                                        header_message=message)

        elif SUBMIT == 'Upload File':
            valid   = 0
            message = "Please specify a a non-empty file to upload."
            if upload_file and upload_file.filename:
                upload_contents = upload_file.read()
                if not upload_contents:
                    message = """Uploaded file %s is empty.
                    Please specify a non-empty file.\n""" % upload_file.filename
                else:
                    valid = 1;

            if valid:
                self.setContents(upload_contents, REQUEST)
                message = 'Content changed. (%s)' % DateTime().ISO()
                return self.manage_editForm(externalfile_content,
                                            header_message=message)
            else:
                message.replace('\n','<br/>')
                return MessageDialog(title   = 'ERROR',
                                     message = message,
                                     action  = 'manage_editForm')

        else: # change size of textarea form control 
            colDelta, rowDelta = self._size_changes[SUBMIT]
            cols=max(self._minCols, string.atoi(edit_textarea_cols)+colDelta)
            rows=max(self._minRows, string.atoi(edit_textarea_rows)+rowDelta)

            e=(DateTime('GMT') + 365).rfc822()

            resp = REQUEST['RESPONSE'] 
            resp.setCookie('edit_textarea_cols',str(cols),path='/',expires=e)
            resp.setCookie('edit_textarea_rows',str(rows),path='/',expires=e)
        
            return self.manage_editForm(self, externalfile_content, REQUEST,
                                        edit_textarea_cols = cols,
                                        edit_textarea_rows = rows,
                                        header_message='Size changed.')

    # Saving the file back via FTP or webdav
    _security.declareProtected('Change External Files', 'PUT')
    def PUT(self, REQUEST, RESPONSE):
        "Handle HTTP PUT requests"
        self.dav__init(REQUEST, RESPONSE)
        instream = REQUEST['BODYFILE']
        self.setContents(instream.read(), REQUEST)
        RESPONSE.setStatus(204)
        return RESPONSE

    #  FTP "get"
    _security.declareProtected('FTP access', 'manage_FTPget')
    manage_FTPget = getContents

    _security.declareProtected('Search ZCatalog', 'PrincipiaSearchSource')
    def PrincipiaSearchSource(self, REQUEST=None):
        "Support for full text search via ZCatalog indexing"
        if self.getDisplayType() == 'ascii':
            return self.getContents(REQUEST)
        else:
            return ""

    ################################################################
    # Support for Schema Migration
    ################################################################
    
    def repair(self):
        """

        Repair this object.  This method is used for schema migration,
        when the class definition changes and existing instances of
        previous versions must be updated.  See 'schema migration'
        discussion in README.txt

        """
        if not hasattr(self, 'behave_like'):
            self.behave_like = 'DTMLDocument'

# register security information
Globals.InitializeClass(ExternalFile)
